#include <glad/glad.h>
#include <GLFW/glfw3.h>
#include <iostream>
#include <fstream>
#include <string>

#include "shaders/Shader.h"

// Continue: https://learnopengl.com/Getting-started/Shaders
// Chapter: Not yet started
//


// 
// HOW TO redirect error output to file
// Maybe it is possible to redirect to another stream object which adds infos like thread id/time etc.?
// 
// std::ofstream errorFile("error.log");
// std::streambuf* originalCerrBuffer = std::cerr.rdbuf();  // Save original buffer
// Redirect cerr to the error file
// std::cerr.rdbuf(errorFile.rdbuf());

#define WINDOW_WIDTH 800
#define WINDOW_HEIGHT 600


void OnWindowResize(GLFWwindow* window, int width, int height)
{
	glViewport(0, 0, width, height);
}

void ProcessInput(GLFWwindow* window)
{
	if (glfwGetKey(window, GLFW_KEY_ESCAPE) == GLFW_PRESS)
		glfwSetWindowShouldClose(window, true);
}

int main()
{
	glfwInit();
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

	GLFWwindow* window = glfwCreateWindow(WINDOW_WIDTH, WINDOW_HEIGHT, "Henlo worl", nullptr, nullptr);

	if (window == nullptr)
	{
		std::cerr << "ERROR::GLFW::CREATE_WINDOW";
		glfwTerminate();

		return -1;
	}

	glfwMakeContextCurrent(window);

	if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
	{
		std::cerr << "ERROR::GLAD::INITILISE";
		return -1;
	}

	OnWindowResize(window, WINDOW_WIDTH, WINDOW_HEIGHT);
	glfwSetFramebufferSizeCallback(window, OnWindowResize);

	GLfloat vertices[] = {
	  0.5f,  0.5f, 0.0f,   1.0f, 0.0f, 0.0f,  // top right
	  0.5f, -0.5f, 0.0f,   0.0f, 1.0f, 0.0f,  // bottom right
	 -0.5f, -0.5f, 0.0f,   0.0f, 0.0f, 1.0f,  // bottom left
	 -0.5f,  0.5f, 0.0f,   1.0f, 0.0f, 0.0f   // top left 
	};
	GLuint faces[]{
		0, 1, 3,
		1, 2, 3
	};

	GLuint vertexBufferObject, elementBufferObject, vertexArrayObject;
	glGenVertexArrays(1, &vertexArrayObject);
	glGenBuffers(1, &vertexBufferObject);
	glGenBuffers(1, &elementBufferObject);

	glBindVertexArray(vertexArrayObject);

	glBindBuffer(GL_ARRAY_BUFFER, vertexBufferObject);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, elementBufferObject);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(faces), faces, GL_STATIC_DRAW);

	// position attribute
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)0);
	glEnableVertexAttribArray(0);
	// colour (or more general, the second vertex Attribute)
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)(3 * sizeof(float)));
	glEnableVertexAttribArray(1);
	// I could unbind the VAO, but this is usually not needed because when creating a new VAO we need to bind that to change it anyway, so there shouldn't be a problem

	std::string vertexPath = "shaders/default/default.vert";
	std::string fragmentPath = "shaders/default/default.frag";

	std::unique_ptr<Nebulix::Shader> shader;
	try
	{
		shader = std::make_unique<Nebulix::Shader>(vertexPath, fragmentPath);
	}
	catch (const std::exception &e)
	{
		std::cerr << e.what();
		return -1;
	}
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	
	// main loop
	while(!glfwWindowShouldClose(window))
	{
		glClearColor(0.2f, 0.3f, 0.3f, 1.0f);
		glClear(GL_COLOR_BUFFER_BIT);

		ProcessInput(window);

		shader->Use();
		//float time = glfwGetTime();
		//float greenValue = (sin(time) / 2.0f) + 0.5f;
		//shader->setFloat("ourColour", 0.0f, greenValue, 0.0f, 0.0f);

		glBindVertexArray(vertexArrayObject);
		glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0);

		glfwSwapBuffers(window);
		glfwPollEvents();
	}

	// optional, as we are at the end of the program.
	glDeleteVertexArrays(1, &vertexArrayObject);
	glDeleteBuffers(1, &elementBufferObject);

	glfwTerminate();
	return 0;
}